from dotmap import DotMap
import yaml
import logging

from . import exceptions

log = logging.getLogger('root')


def load_YAML(file_object):
    """
    Load a YAML configuration file into a DotMap.
    """
    try:
        config = yaml.safe_load(file_object)
    except yaml.composer.ComposerError as e:
        log.error(f"YAML composer error {e}")
        raise exceptions.ConfigurationError(e) from e
    except yaml.scanner.ScannerError as e:
        log.error(f"YAML scanner error {e}")
        raise exceptions.ConfigurationError(e) from e

    return DotMap(config, _dynamic=True)


def make_static(dm: DotMap):
    """
    Recursively remove _dynamic from a DotMap.
    """
    if isinstance(dm, DotMap):
        for k in dm._map:
            if isinstance(dm[k], DotMap):
                make_static(dm[k])
            if isinstance(dm[k], list):
                for item in dm[k]:
                    make_static(item)
            dm._dynamic = False



import argparse
from pathlib import Path


class ReadableDir(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        path = Path(values)
        if not path.is_dir():
            raise argparse.ArgumentTypeError(f"ReadableDir: {path} is not a valid path")
        try:
            setattr(namespace, self.dest, path)
        except IOError:
            raise argparse.ArgumentTypeError(f"ReadableDir: {path} is not a readable directory")


class WriteableDir(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        path = Path(values)
        path.mkdir(parents=True, exist_ok=True)

        if not path.is_dir():
            raise argparse.ArgumentTypeError(f"WriteableDir: {path} is not a valid path")
        try:
            setattr(namespace, self.dest, path)
        except IOError:
            raise argparse.ArgumentTypeError(f"WriteableDir: {path} is not a writeable directory")



import abc
import argparse
import dotmap
import logging
import sys

import schema

from . import scala
from . import exceptions
from . import configuration
from . import colour as c


log = logging.getLogger('root')


class Scalyca(scala.Scala, metaclass=abc.ABCMeta):
    """
    Simple Console Application with Logging, Yaml Configuration and Argparse

    Scalyca extends Scala with support for persistent configuration that is read from a YAML file.
    An extra mandatory argument `config` is added and the file is parsed automatically.
    Its contents are stored in a DotMap, so that they can be accessed more conveniently.

    In `override_configuration` it is possible to override values from the configuration file
    with command line arguments.
    """
    _app_name = 'Scalyca'
    _description = "Simple Configurable Application with Logging, Yaml file Configuration and Argparse"
    # Config validation schema, or None to skip validation
    _prog = "Scalyca"
    _schema: schema.Schema | None = None

    def _add_default_arguments(self):
        """
        Scalyca adds one mandatory argument to Scala's defaults: the path to the YAML config file.
        """
        super()._add_default_arguments()
        self.add_argument('config', type=argparse.FileType('r'), help="main configuration file")

    def _configure(self):
        try:
            self._load_configuration()
            self.override_configuration()
            self._lock_configuration()
            self._debug_configuration()
            self._validate_configuration()

        except exceptions.FatalError:
            log.critical(f"{c.script(self._app_name)} aborting during configuration")
            sys.exit(-1)

    def _load_configuration(self):
        """ Load the configuration from the specified file and convert it to a dynamic dotmap """
        raw = configuration.load_YAML(self.args.config)
        self.config = dotmap.DotMap(raw, _dynamic=True)
        log.debug(f"Configuration read from {c.path(self.args.config.name)}")

    def _lock_configuration(self):
        """
        Lock the configuration (set the DotMap's ``_dynamic`` to false recursively).
        """
        log.debug(f"Locking the configuration")
        configuration.make_static(self.config)

    def _debug_configuration(self):
        if self.args.debug:
            log.debug("Full config is")
            self.config.pprint()

    def _validate_configuration(self):
        if self._schema is None:
            log.debug("No config validation schema provided, skipping validation")
        else:
            log.debug("Validating the configuration")
            try:
                self._schema.validate(self.config)
            except schema.SchemaError as e:
                raise exceptions.ConfigurationError(e) from e



import colorama
from colorama import Fore, Style

colorama.init()


def colour(what, how):
    return f"{how}{what}{Style.RESET_ALL}"


def ok(what):
    return colour(what, Fore.GREEN)


def num(what):
    return colour(what, Fore.CYAN)


def act(what):
    return colour(what, Fore.YELLOW)


def warn(what):
    return colour(what, Fore.YELLOW)


def err(what):
    return colour(what, Fore.RED)


def critical(what):
    return colour(what, Fore.RED)


def path(what):
    return colour(what, Fore.LIGHTMAGENTA_EX)


def name(what):
    return colour(what, Fore.YELLOW)


def over(what):
    return colour(what, Fore.LIGHTGREEN_EX)


def script(what):
    return colour(what, Fore.LIGHTGREEN_EX)


def param(what):
    return colour(what, Fore.LIGHTBLUE_EX)


def debug(what):
    return colour(what, Fore.LIGHTBLACK_EX)


def none(what):
    return colour(what, Fore.WHITE)



import abc
import argparse
import datetime
import logging

from . import colour as c
from . import exceptions
from . import logger

log = logger.setup_log('root', timefmt='%H:%M:%S')


class Scala(metaclass=abc.ABCMeta):
    """
    Scala: Simple Console Application with Logging and Argparse

    The main class of the framework that manages basic application metadata,
    setting up the logging module and parsing command-line arguments.
    It is intended for batch processing jobs and other non-interactive programs.

    Its primary purpose it to make parsing arguments straightforward and simple,
    and provide defaults for the `logging` module.
    Basic errors are also handled relatively gracefully.

    In order to use it fo your simple console application, you need to:

    0.  Set up `_description`, `_prog` and `_version` (optional)
    1.  Provide a body for ``add_arguments`` (optional).
        Several default arguments, like `--logfile` and `--debug` are added automatically.
    2.  Provide a body for `override_configuration`.
        More
        (for instance more complex argument checking that is not possible to do with `argparse`).
    3.  Provide a body for `initialize` (optional).
        Internal attributes (like processing the config and CLI arguments) should be done in this step.
        For instance, converting `datetime` objects to `astropy.Time`.
    4.  Provide a body for `main` (this one is required).
        The main body of your program should go there.
    5.  Provide a body for `finalize` (optional).
        Cleanup and output of a summary report should go there.
    """
    # Verbose description of the application
    _description = "Simple Configurable Application with Logging and Argparse"
    # Prog, a short name of the program that is shown in --help
    _prog = "Scala"
    # Program version, also shown in --help
    _version = "undefined"

    def __init__(self, **kwargs):
        """
        Optionally override the application name and description.

        prog:
            override the short application name
        description:
            override the short application description
        success_message:
            override the final success message
        abort_message:
            override the program crash message
        """
        self._ok = False
        self._started = datetime.datetime.now(datetime.UTC)
        self._description = kwargs.get('description', self._description)
        self._prog = kwargs.get('prog', self._prog)
        self._success_message = kwargs.get('success_message', f"{c.script(self._prog)} finished successfully")
        self._abort_message = kwargs.get('error_message', f"{c.script(self._prog)} aborted during runtime")

        self._argparser = argparse.ArgumentParser(
            prog=self._prog,
            description=self._description
        )

    def _initialize(self):
        log.info(f"{c.script(self._prog)} (version {self._version}) starting")
        self._add_default_arguments()
        self.add_arguments()
        self.args = self._argparser.parse_args()

        self._configure()
        self._override_configuration()
        self.initialize()

    def add_argument(self, *args, **kwargs):
        """
        Just a public wrapper so that we do not have to access _argparser directly.
        All arguments *args and **kwargs are directly passed to internal argparse.ArgumentParser.
        """
        self._argparser.add_argument(*args, **kwargs)

    def add_arguments(self):
        """ Override this in your Scala derivative to add custom arguments """

    def _add_default_arguments(self):
        """
        Add default arguments that are present in every instance of Scala.
        """
        self.add_argument('-l', '--logfile', type=argparse.FileType('w'), help="Write log to file")
        self.add_argument('-d', '--debug', action='store_true', help="Turn on verbose logging")

    def initialize(self):
        """ Custom initialization routines. Might remain empty. """

    def _configure(self):
        pass

    @abc.abstractmethod
    def main(self):
        """
        The main entry point of the program.
        Provide an implementation in your derivative of Scala.
        """

    def finalize(self):
        """ Override this in you Scala derivative to add custom finalization """

    def _override_configuration(self):
        self.override_configuration()
        log.setLevel(logging.DEBUG if self.args.debug else logging.INFO)
        log.debug(f"{c.script(self._prog)} debug mode on")

        if self.args.logfile:
            log.addHandler(logging.FileHandler(self.args.logfile.name))
            log.debug(f"Added log output {c.path(self.args.logfile.name)}")

    def override_configuration(self):
        """ Custom override hook for overriding the configuration """

    def run(self):
        try:
            self._initialize()
            self.main()
            self._ok = True
        except exceptions.PrerequisiteError as e:
            log.error(f"Terminating due to missing prerequisites: {e}")
            self._ok = False
        except exceptions.ConfigurationError as e:
            log.error(f"Terminating due to a configuration error: {e}")
            self._ok = False
        except KeyboardInterrupt:
            log.error(f"Interrupted by user")
            self._ok = False
        finally:
            self._finalize()
            run_time = datetime.datetime.now(datetime.UTC) - self._started
            if self._ok:
                log.info(f"{self._success_message} in {run_time}")
            else:
                log.critical(f"{self._abort_message} after {run_time}")

    def _finalize(self):
        self.finalize()



class ConfigurationError(BaseException):
    pass

class OverwriteError(Exception):
    pass

class PrerequisiteError(Exception):
    pass

class CommandLineError(Exception):
    pass

class FatalError(Exception):
    pass



import logging
import time
import sys

from . import colour as c


class Formatter(logging.Formatter):
    def format(self, record) -> str:
        record.level = {
            'DEBUG':    c.debug,
            'INFO':     c.ok,
            'WARNING':  c.warn,
            'ERROR':    c.err,
            'CRITICAL': c.critical,
        }[record.levelname](record.levelname[:3])
        return super().format(record)

    def formatTime(self, record, datefmt) -> str:
        ct = self.converter(record.created)
        return f"{time.strftime(datefmt, ct)}.{int(record.msecs):03d}"


class ShortFormatter(Formatter):
    def format(self, record) -> str:
        record.asctime = {
            'DEBUG':    c.debug,
            'INFO':     c.ok,
            'WARNING':  c.warn,
            'ERROR':    c.err,
            'CRITICAL': c.critical,
        }[record.levelname](record.created)
        return super().format(record)


def setup_log(name, *, output=None, fmt='[{asctime} {level}] {message}', timefmt='%Y-%m-%d %H:%M:%S'):
    formatter = ShortFormatter('[{asctime}] {message}', timefmt, '{')

    if isinstance(output, str):
        handler = logging.FileHandler(output)
    else:
        handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(formatter)

    log = logging.getLogger(name)
    log.setLevel(logging.INFO)
    log.addHandler(handler)

    return log
